# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Records/Drafts table generator tests."""

from copy import deepcopy
from unittest.mock import patch

from invenio_rdm_migrator.load.ids import pid_pk
from invenio_rdm_migrator.streams.models.communities import RDMParentCommunityMetadata
from invenio_rdm_migrator.streams.models.pids import PersistentIdentifier
from invenio_rdm_migrator.streams.models.records import (
    RDMDraftMetadata,
    RDMParentMetadata,
    RDMRecordMetadata,
)
from invenio_rdm_migrator.streams.records.load import (
    RDMDraftTableGenerator,
    RDMRecordTableGenerator,
)

# IMPORTANT NOTE: since the resolve references method would be called externally
# (i.e. not by generate_rows) not all pid pk have consistent values coming from
# pid_pk(). Therefore, some calls are made at the beginning of the tests.


class MockUTCDate:
    """Mock UTC date."""

    def isoformat(self):
        """ISO formatter datetime."""
        return "2023-04-01 12:00:00.00000"


class MockDateTime:
    """Mock datetime class."""

    def utcnow(self):
        """Nop func."""
        return MockUTCDate()


@patch(
    "invenio_rdm_migrator.streams.records.table_generators.records.datetime",
    MockDateTime(),
)
@patch(
    "invenio_rdm_migrator.streams.records.table_generators.parents.datetime",
    MockDateTime(),
)
def test_single_record_generate_rows(
    state, communities_state, parents_state, transformed_record_entry_pks
):
    """A published record with a non state parent.

    It does not make sense to also test with a state parent since that would mean
    there was a previous version of the record, which is tested separately in this module.
    """
    pid_1 = pid_pk()
    pid_2 = pid_pk()
    tg = RDMRecordTableGenerator()
    rows = list(tg._generate_rows(transformed_record_entry_pks))
    expected_rows = [
        PersistentIdentifier(  # parent recid
            id=pid_1,
            pid_type="recid",
            pid_value="12345677",
            status="R",
            object_type="rec",
            object_uuid="12345678-abcd-1a2b-3c4d-123abc456def",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
        RDMParentMetadata(  # parent record
            id="12345678-abcd-1a2b-3c4d-123abc456def",
            json={
                "id": "12345677",
                "pid": {
                    "pk": pid_1,
                    "obj_type": "rec",
                    "pid_type": "recid",
                    "status": "R",
                },
                "communities": {
                    "ids": [
                        "12345678-abcd-1a2b-3c4d-123abc456def",
                        "12345678-abcd-1a2b-3c4d-123abc123abc",
                    ],
                    "default": "12345678-abcd-1a2b-3c4d-123abc456def",
                },
            },
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            version_id=1,
        ),
        RDMParentCommunityMetadata(  # parent community
            community_id="12345678-abcd-1a2b-3c4d-123abc456def",
            record_id="12345678-abcd-1a2b-3c4d-123abc456def",
            request_id=None,
        ),
        RDMRecordMetadata(  # record
            id="2d6970ea-602d-4e8b-a918-063a59823386",
            json={
                "id": "12345678",
                "pid": {
                    "pk": pid_2,
                    "obj_type": "rec",
                    "pid_type": "recid",
                    "status": "R",
                },
                "pids": {
                    "oai": {
                        "provider": "oai",
                        "identifier": "oai:zenodo.org:12345678",
                    },
                    "doi": {
                        "client": "datacite",
                        "provider": "datacite",
                        "identifier": "10.5281/zenodo.12345678",
                    },
                },
            },
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            version_id=1,
            index=1,
            bucket_id="bur3c0rd-1234-abcd-1ab2-1234abcd56ef",
            media_bucket_id=None,
            parent_id="12345678-abcd-1a2b-3c4d-123abc456def",
            deletion_status="P",
        ),
        PersistentIdentifier(  # recid
            id=pid_2,
            pid_type="recid",
            pid_value="12345678",
            status="R",
            object_type="rec",
            object_uuid="2d6970ea-602d-4e8b-a918-063a59823386",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
        PersistentIdentifier(  # doi
            id=1_000_002,
            pid_type="doi",
            pid_value="10.5281/zenodo.12345678",
            status="R",
            object_type="rec",
            object_uuid="2d6970ea-602d-4e8b-a918-063a59823386",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
        PersistentIdentifier(  # oai
            id=1_000_003,
            pid_type="oai",
            pid_value="oai:zenodo.org:12345678",
            status="R",
            object_type="rec",
            object_uuid="2d6970ea-602d-4e8b-a918-063a59823386",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
    ]

    assert rows == expected_rows

    assert len(list(state.PARENTS.all())) == 2  # pre-existing and new
    assert len(list(state.RECORDS.all())) == 1


@patch(
    "invenio_rdm_migrator.streams.records.table_generators.drafts.datetime",
    MockDateTime(),
)
@patch(
    "invenio_rdm_migrator.streams.records.table_generators.parents.datetime",
    MockDateTime(),
)
def test_single_draft_generate_rows(
    state, communities_state, parents_state, transformed_draft_entry_pks
):
    """A new draft, not published."""
    pid_1 = pid_pk()
    pid_2 = pid_pk()
    tg = RDMDraftTableGenerator()
    rows = list(tg._generate_rows(transformed_draft_entry_pks))
    expected_rows = [
        PersistentIdentifier(  # parent recid
            id=pid_1,
            pid_type="recid",
            pid_value="12345677",
            status="N",
            object_type="rec",
            object_uuid="12345678-abcd-1a2b-3c4d-123abc456def",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
        RDMParentMetadata(  # parent record
            id="12345678-abcd-1a2b-3c4d-123abc456def",
            json={
                "id": "12345677",
                "pid": {
                    "pk": pid_1,
                    "obj_type": "rec",
                    "pid_type": "recid",
                    "status": "N",
                },
                "communities": {
                    "ids": [
                        "12345678-abcd-1a2b-3c4d-123abc456def",
                        "12345678-abcd-1a2b-3c4d-123abc123abc",
                    ],
                    "default": "12345678-abcd-1a2b-3c4d-123abc456def",
                },
            },
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            version_id=1,
        ),
        RDMDraftMetadata(  # record
            id="2d6970ea-602d-4e8b-a918-063a59823386",
            json={
                "id": "12345678",
                "pid": {
                    "pk": pid_2,
                    "obj_type": "rec",
                    "pid_type": "recid",
                    "status": "N",
                },
                "pids": {
                    "doi": {
                        "provider": "external",
                        "identifier": "10.1234/foo",
                    },
                },
            },
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            version_id=1,
            index=1,
            bucket_id="b0c73700-1234-abcd-1ab2-1234abcd56ef",
            media_bucket_id=None,
            parent_id="12345678-abcd-1a2b-3c4d-123abc456def",
            expires_at="2024-01-01 12:00:00.00000",
            fork_version_id=None,
        ),
        PersistentIdentifier(  # recid
            id=pid_2,
            pid_type="recid",
            pid_value="12345678",
            status="N",
            object_type="rec",
            object_uuid="2d6970ea-602d-4e8b-a918-063a59823386",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
    ]

    assert rows == expected_rows

    assert len(list(state.PARENTS.all())) == 2  # pre-existing and new
    assert len(list(state.RECORDS.all())) == 0


@patch(
    "invenio_rdm_migrator.streams.records.table_generators.records.datetime",
    MockDateTime(),
)
@patch(
    "invenio_rdm_migrator.streams.records.table_generators.drafts.datetime",
    MockDateTime(),
)
@patch(
    "invenio_rdm_migrator.streams.records.table_generators.parents.datetime",
    MockDateTime(),
)
def test_record_versions_and_old_draft_generate_rows(
    state,
    communities_state,
    parents_state,
    transformed_record_entry_pks,
    transformed_draft_entry_pks,
):
    """A record with two versions (v1, v2) and a draft of the first version (v1)."""
    tgs = [
        RDMRecordTableGenerator(),
        RDMDraftTableGenerator(),
    ]

    v1 = transformed_record_entry_pks
    v2 = deepcopy(transformed_record_entry_pks)
    v2["record"]["id"] = "2d6970ea-602d-4e8b-a918-063a59823387"
    v2["record"]["index"] = 2
    v2["record"]["json"]["id"] = "12345679"
    v2["record"]["json"]["pid"]["pk"] = 1_000_002  # pk, not pid value
    v2["record"]["json"]["pids"]["oai"]["identifier"] = "oai:zenodo.org:12345679"
    v2["record"]["json"]["pids"]["doi"]["identifier"] = "10.5281/zenodo.12345679"
    d_v1 = transformed_draft_entry_pks
    d_v1["draft"]["json"]["pid"]["pk"] = 1_000_002  # pk, not pid value, same as v1
    d_v1["draft"]["json"]["pid"][
        "status"
    ] = "R"  # same as already R pid form the record
    d_v1["draft"]["fork_version_id"] = 1

    rows = []
    for entry in [v1, v2, d_v1]:
        for tg in tgs:
            rows.extend(list(tg._generate_rows(entry)))

    expected_rows_v2 = [
        RDMRecordMetadata(  # record
            id="2d6970ea-602d-4e8b-a918-063a59823387",
            json={
                "id": "12345679",
                "pid": {
                    "pk": 1_000_002,
                    "obj_type": "rec",
                    "pid_type": "recid",
                    "status": "R",
                },
                "pids": {
                    "oai": {
                        "provider": "oai",
                        "identifier": "oai:zenodo.org:12345679",
                    },
                    "doi": {
                        "client": "datacite",
                        "provider": "datacite",
                        "identifier": "10.5281/zenodo.12345679",
                    },
                },
            },
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            version_id=1,
            index=2,
            bucket_id="bur3c0rd-1234-abcd-1ab2-1234abcd56ef",
            media_bucket_id=None,
            parent_id="12345678-abcd-1a2b-3c4d-123abc456def",
            deletion_status="P",
        ),
        PersistentIdentifier(  # recid
            id=1_000_002,  # this called is mocked and will not increment the counter
            pid_type="recid",
            pid_value="12345679",
            status="R",
            object_type="rec",
            object_uuid="2d6970ea-602d-4e8b-a918-063a59823387",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
        PersistentIdentifier(  # doi
            id=1_000_002,
            pid_type="doi",
            pid_value="10.5281/zenodo.12345679",
            status="R",
            object_type="rec",
            object_uuid="2d6970ea-602d-4e8b-a918-063a59823387",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
        PersistentIdentifier(  # oai
            id=1_000_003,
            pid_type="oai",
            pid_value="oai:zenodo.org:12345679",
            status="R",
            object_type="rec",
            object_uuid="2d6970ea-602d-4e8b-a918-063a59823387",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
    ]

    expected_rows_d_v1 = [
        RDMDraftMetadata(  # record
            id="2d6970ea-602d-4e8b-a918-063a59823386",
            json={
                "id": "12345678",
                "pid": {
                    "pk": 1_000_002,
                    "obj_type": "rec",
                    "pid_type": "recid",
                    "status": "R",
                },
                "pids": {
                    "doi": {
                        "provider": "external",
                        "identifier": "10.1234/foo",
                    },
                },
            },
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            version_id=1,
            index=1,
            bucket_id="b0c73700-1234-abcd-1ab2-1234abcd56ef",
            media_bucket_id=None,
            parent_id="12345678-abcd-1a2b-3c4d-123abc456def",
            expires_at="2024-01-01 12:00:00.00000",
            fork_version_id=1,
        ),
    ]

    assert len(rows) == 12
    # v1 rows not asserted since they are checked at test_single_record_generate_rows
    assert rows[7:11] == expected_rows_v2
    assert rows[11:12] == expected_rows_d_v1

    assert len(list(state.PARENTS.all())) == 2  # pre-existing and new
    assert len(list(state.RECORDS.all())) == 2  # two added records


@patch(
    "invenio_rdm_migrator.streams.records.table_generators.records.datetime",
    MockDateTime(),
)
@patch(
    "invenio_rdm_migrator.streams.records.table_generators.drafts.datetime",
    MockDateTime(),
)
@patch(
    "invenio_rdm_migrator.streams.records.table_generators.parents.datetime",
    MockDateTime(),
)
def test_record_and_new_version_draft_generate_rows(
    state,
    communities_state,
    parents_state,
    transformed_record_entry_pks,
    transformed_draft_entry_pks,
):
    """A published record (v1) with a new version draft (v2)."""
    tgs = [
        RDMRecordTableGenerator(),
        RDMDraftTableGenerator(),
    ]

    pid_1 = pid_pk()
    pid_2 = pid_pk()
    v1 = transformed_record_entry_pks
    v2 = deepcopy(transformed_record_entry_pks)
    v2["record"]["id"] = "2d6970ea-602d-4e8b-a918-063a59823387"
    v2["record"]["index"] = 2
    v2["record"]["json"]["id"] = "12345679"
    v2["record"]["json"]["pid"]["pk"] = pid_1  # pk, not pid value
    v2["record"]["json"]["pids"]["oai"]["identifier"] = "oai:zenodo.org:12345679"
    v2["record"]["json"]["pids"]["doi"]["identifier"] = "10.5281/zenodo.12345679"
    d_v3 = transformed_draft_entry_pks
    d_v3["draft"]["id"] = "2d6970ea-602d-4e8b-a918-063a59823389"
    d_v3["draft"]["json"]["id"] = "12345680"
    d_v3["draft"]["json"]["pid"]["pk"] = pid_2  # pk, not pid value, same as v1
    d_v3["draft"]["json"]["pids"]["doi"]["identifier"] = "10.1234/bar"

    rows = []
    for entry in [v1, v2, d_v3]:
        for tg in tgs:
            rows.extend(list(tg._generate_rows(entry)))

    expected_rows_d_v3 = [
        RDMDraftMetadata(  # record
            id="2d6970ea-602d-4e8b-a918-063a59823389",
            json={
                "id": "12345680",
                "pid": {
                    "pk": pid_2,
                    "obj_type": "rec",
                    "pid_type": "recid",
                    "status": "N",
                },
                "pids": {
                    "doi": {
                        "provider": "external",
                        "identifier": "10.1234/bar",
                    },
                },
            },
            created="2023-01-01 12:00:00.00000",
            updated="2023-01-31 12:00:00.00000",
            version_id=1,
            index=1,
            bucket_id="b0c73700-1234-abcd-1ab2-1234abcd56ef",
            media_bucket_id=None,
            parent_id="12345678-abcd-1a2b-3c4d-123abc456def",
            expires_at="2024-01-01 12:00:00.00000",
            fork_version_id=None,
        ),
        PersistentIdentifier(  # recid
            id=pid_2,  # this called is mocked and will not increment the counter
            pid_type="recid",
            pid_value="12345680",
            status="N",
            object_type="rec",
            object_uuid="2d6970ea-602d-4e8b-a918-063a59823389",
            created="2023-04-01 12:00:00.00000",
            updated="2023-04-01 12:00:00.00000",
        ),
    ]

    assert len(rows) == 13

    # v1 rows not asserted since they are checked at test_single_record_generate_rows
    # v2 rows not asserted since they are checked at test_record_versions_and_old_draft_generate_rows
    assert rows[11:13] == expected_rows_d_v3

    assert len(list(state.PARENTS.all())) == 2  # pre-existing and new
    assert len(list(state.RECORDS.all())) == 2  # two added records



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration user transform interfaces."""

from abc import abstractmethod

from ...transform import Entry, Transform


class FilesTransform(Transform):
    """Base class for user data transformation."""

    @abstractmethod
    def _bucket(self, entry):
        """Transform the bucket."""
        pass

    @abstractmethod
    def _object_version(self, entry):
        """Transform the object_version."""
        pass

    @abstractmethod
    def _file(self, entry):
        """Transform the file."""
        pass

    def _transform(self, entry):
        """Transform a single entry."""
        return {
            "bucket": self._bucket(entry),
            "object_version": self._object_version(entry),
            "file": self._file(entry),
        }


class FilesBucketEntry(Entry):
    """Transform a single file bucket entry."""

    @abstractmethod
    def _id(self, entry):
        """Returns the file bucket ID."""
        pass

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _default_location(self, entry):
        """Returns the file bucket default location."""
        pass

    @abstractmethod
    def _default_storage_class(self, entry):
        """Returns the file bucket default storage class."""
        pass

    @abstractmethod
    def _size(self, entry):
        """Returns the file bucket size."""
        pass

    @abstractmethod
    def _quota_size(self, entry):
        """Returns the file bucket quota size."""
        pass

    @abstractmethod
    def _max_file_size(self, entry):
        """Returns the file bucket max file size."""
        pass

    @abstractmethod
    def _locked(self, entry):
        """Returns if the file bucket is locked."""
        pass

    @abstractmethod
    def _deleted(self, entry):
        """Returns if the file bucket is deleted."""
        pass

    def transform(self, entry):
        """Transform a user single entry."""
        return {
            "id": self._id(entry),
            "created": self._created(entry),
            "updated": self._updated(entry),
            "default_location": self._default_location(entry),
            "default_storage_class": self._default_storage_class(entry),
            "size": self._size(entry),
            "quota_size": self._quota_size(entry),
            "max_file_size": self._max_file_size(entry),
            "locked": self._locked(entry),
            "deleted": self._deleted(entry),
        }


class FilesObjectVersionEntry(Entry):
    """Transform a single file object version entry."""

    @abstractmethod
    def _version_id(self, entry):
        """Returns the file object version ID."""
        pass

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _key(self, entry):
        """Returns the file key name."""
        pass

    @abstractmethod
    def _bucket_id(self, entry):
        """Returns the file object version bucket ID i.e associated bucket."""
        pass

    @abstractmethod
    def _file_id(self, entry):
        """Returns the file object version file ID i.e associated file."""
        pass

    @abstractmethod
    def _mimetype(self, entry):
        """Returns the file object version mimetype."""
        pass

    @abstractmethod
    def _is_head(self, entry):
        """Returns if the file object version is head i.e the latest."""
        pass

    def transform(self, entry):
        """Transform a user single entry."""
        return {
            "version_id": self._version_id(entry),
            "created": self._created(entry),
            "updated": self._updated(entry),
            "key": self._key(entry),
            "bucket_id": self._bucket_id(entry),
            "file_id": self._file_id(entry),
            "_mimetype": self._mimetype(entry),
            "is_head": self._is_head(entry),
        }


class FilesInstanceEntry(Entry):
    """Transform a single file instance entry."""

    @abstractmethod
    def _id(self, entry):
        """Returns the file instance ID."""
        pass

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _uri(self, entry):
        """Returns the file uri."""
        pass

    @abstractmethod
    def _storage_class(self, entry):
        """Returns the file instance storage class."""
        pass

    @abstractmethod
    def _size(self, entry):
        """Returns the file instance size."""
        pass

    @abstractmethod
    def _checksum(self, entry):
        """Returns the file instance checksum."""
        pass

    @abstractmethod
    def _readable(self, entry):
        """Returns if the file instance can be read."""
        pass

    @abstractmethod
    def _writable(self, entry):
        """Returns if the file instance can be written."""
        pass

    @abstractmethod
    def _last_check_at(self, entry):
        """Returns the last date the file was checked."""
        pass

    @abstractmethod
    def _last_check(self, entry):
        """Returns if the file instance was last checked."""
        pass

    def transform(self, entry):
        """Transform a user single entry."""
        return {
            "id": self._id(entry),
            "created": self._created(entry),
            "updated": self._updated(entry),
            "uri": self._uri(entry),
            "storage_class": self._storage_class(entry),
            "size": self._size(entry),
            "checksum": self._checksum(entry),
            "readable": self._readable(entry),
            "writable": self._writable(entry),
            "last_check_at": self._last_check_at(entry),
            "last_check": self._last_check(entry),
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""PostgreSQL sequence handling."""


import psycopg


class AlterSequencesMixin:
    """Mixin to update sequences to the latest value."""

    def alter_sequences(self):
        """Query all sequences and update those that were modified."""
        tables = set()
        for tg in self.table_generators:
            tables = tables.union(set(tg.tables))
            tg.post_load(db_uri=self.db_uri)

        with psycopg.connect(self.db_uri) as conn:
            sequences = conn.execute(
                """
                SELECT
                    t.oid::regclass AS table_name,
                    a.attname AS column_name,
                    s.relname AS sequence_name
                FROM pg_class AS t
                    JOIN pg_attribute AS a ON a.attrelid = t.oid
                    JOIN pg_depend AS d ON d.refobjid = t.oid AND d.refobjsubid = a.attnum
                    JOIN pg_class AS s ON s.oid = d.objid
                WHERE
                    d.classid = 'pg_catalog.pg_class'::regclass
                    AND d.refclassid = 'pg_catalog.pg_class'::regclass
                    AND d.deptype IN ('i', 'a')
                    AND t.relkind IN ('r', 'P')
                    AND s.relkind = 'S';
                """
            )

            for seq in sequences:
                table_name, column, seq_name = seq
                if table_name in tables:
                    max_val = conn.execute(f"SELECT MAX({column}) FROM {table_name}")
                    max_val = list(max_val)[0][0]  # get actual value from iterator
                    if max_val:  # if no updates it returns None
                        conn.execute(
                            f"ALTER SEQUENCE {seq_name} RESTART WITH {max_val}"
                        )
                        # does not require commit as it is a ddl op



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration user transform interfaces."""

from abc import abstractmethod

from ...transform import Entry, Transform


class CommunityTransform(Transform):
    """Base class for community data transformation."""

    @abstractmethod
    def _community(self, entry):
        """Transform the community."""
        pass

    @abstractmethod
    def _oai_set(self, entry):
        """Transform the community OAISet."""
        pass

    @abstractmethod
    def _community_members(self, entry):
        """Transform the community members."""
        pass

    @abstractmethod
    def _featured_community(self, entry):
        """Transform the featured community."""
        pass

    @abstractmethod
    def _community_files(self, entry):
        """Transform the community files."""
        pass

    def _transform(self, entry):
        """Transform a single entry."""
        return {
            "community": self._community(entry),
            "oai_set": self._oai_set(entry),
            "community_files": self._community_files(entry),
            "community_members": self._community_members(entry),
            "featured_community": self._featured_community(entry),
        }


class CommunityEntry(Entry):
    """Transform a single community entry."""

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _version_id(self, entry):
        """Returns the version id.

        {Elastic,Open}search will fail with values lower than 1.
        """
        pass

    @abstractmethod
    def _slug(self, entry):
        """Returns the community slug."""
        pass

    @abstractmethod
    def _files(self, entry):
        """Transform the files of a record."""
        pass

    @abstractmethod
    def _access(self, entry):
        """Transform the access of a record."""
        pass

    @abstractmethod
    def _bucket_id(self, entry):
        """Returns the community bucket id ."""
        pass

    @abstractmethod
    def _metadata(self, entry):
        """Transform the metadata of a community.

        No schema/dictionary is enforced as choice balanced between
        duplicating the schema with abstract methods and allowing for flexibility
        i.e. different metadata schema.
        """
        pass

    def transform(self, entry):
        """Transform a user single entry."""
        return {
            "created": self._created(entry),
            "updated": self._updated(entry),
            "version_id": self._version_id(entry),
            "slug": self._slug(entry),
            "json": {
                "files": self._files(entry),
                "access": self._access(entry),
                "metadata": self._metadata(entry),
            },
            "bucket_id": self._bucket_id(entry),
        }


class CommunityMemberEntry(Entry):
    """Transform a community members."""

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _version_id(self, entry):
        """Returns the version id.

        {Elastic,Open}search will fail with values lower than 1.
        """
        pass

    @abstractmethod
    def _role(self, entry):
        """Returns the community member role."""
        pass

    @abstractmethod
    def _visible(self, entry):
        """Returns if the community member is visible or not."""
        pass

    @abstractmethod
    def _active(self, entry):
        """Returns if the community member is active or not."""
        pass

    @abstractmethod
    def _user_id(self, entry):
        """Returns the community member user id, if the member is a single user."""
        pass

    @abstractmethod
    def _group_id(self, entry):
        """Returns the community member group id, if the member is a group."""
        pass

    @abstractmethod
    def _request_id(self, entry):
        """Returns the community member request id."""
        pass

    def transform(self, entry):
        """Transform a user single entry."""
        return {
            "created": self._created(entry),
            "updated": self._updated(entry),
            "json": {},
            "version_id": self._version_id(entry),
            "role": self._role(entry),
            "visible": self._visible(entry),
            "active": self._active(entry),
            "user_id": self._user_id(entry),
            "group_id": self._group_id(entry),
            "request_id": self._request_id(entry),
        }


class ParentCommunityEntry(Entry):
    """Transform a single parent community entry."""

    @abstractmethod
    def _community(self, entry):
        """Returns the community slug."""
        pass

    @abstractmethod
    def _record(self, entry):
        """Returns the record pid."""
        pass

    @abstractmethod
    def _request(self, entry):
        """Returns the request id."""
        pass

    def transform(self, entry):
        """Transform a record single entry."""
        return {
            "community": self._community(entry),
            "record": self._record(entry),
            "request": self._request(entry),
        }


class FeaturedCommunityEntry(Entry):
    """Transform a single featured community entry."""

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _start_date(self, entry):
        """Returns the start date."""
        pass

    def transform(self, entry):
        """Transform a featured community entry."""
        return {
            "created": self._created(entry),
            "updated": self._updated(entry),
            "start_date": self._start_date(entry),
        }


class CommunityFileEntry(Entry):
    """Transform a community file entry."""

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _id(self, entry):
        """Returns the file id."""
        pass

    @abstractmethod
    def _json(self, entry):
        """Returns the file metadata."""
        pass

    @abstractmethod
    def _version_id(self, entry):
        """Returns the version id."""
        pass

    @abstractmethod
    def _key(self, entry):
        """Return the filepath."""
        pass

    def transform(self, entry):
        """Transforms a community file entry."""
        return {
            "id": self._id(entry),
            "json": self._json(entry),
            "version_id": self._version_id(entry),
            "key": self._key(entry),
            "created": self._created(entry),
            "updated": self._updated(entry),
        }


class OAISetEntry(Entry):
    """Community OAI Set entry."""

    @abstractmethod
    def _id(self, entry):
        """Return the id."""
        pass

    @abstractmethod
    def _spec(self, entry):
        """Return the spec."""
        pass

    @abstractmethod
    def _name(self, entry):
        """Return the name."""
        pass

    @abstractmethod
    def _description(self, entry):
        """Return the description."""
        pass

    @abstractmethod
    def _search_pattern(self, entry):
        """Return the search pattern."""
        pass

    @abstractmethod
    def _system_created(self, entry):
        """Return if set is created by the system."""
        pass

    @abstractmethod
    def _created(self, entry):
        """Return the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Return the update date."""
        pass

    def transform(self, entry):
        """Transforms a community file entry."""
        return {
            "id": self._id(entry),
            "spec": self._spec(entry),
            "name": self._name(entry),
            "description": self._description(entry),
            "search_pattern": self._search_pattern(entry),
            "system_created": self._system_created(entry),
            "created": self._created(entry),
            "updated": self._updated(entry),
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""PostgreSQL Execute load."""

from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from ....logging import Logger
from ...base import Load
from .operations import OperationType


class PostgreSQLTx(Load):
    """PostgreSQL COPY load."""

    def __init__(self, db_uri, _session=None, **kwargs):
        """Constructor."""
        self.db_uri = db_uri
        self._session = _session

    @property
    def _engine(self):
        if self._engine is None:
            self._engine = create_engine(self.db_uri)
        return self._engine

    @property
    def session(self):
        """DB session."""
        if self._session is None:
            self._session = Session(bind=self._engine)
        return self._session

    def _cleanup(self, db=False):
        """No cleanup."""
        logger = Logger.get_logger()
        logger.debug("PostgreSQLExecute does not implement _cleanup()")
        pass

    def _get_obj_by_pk(self, model, data):
        """Get an object based on the primary key."""
        # this function accesses many private methods, variables, assumes indexes, etc.
        # pragmatic implementation, feel free to refactor.
        pk = {}
        for key in model.__mapper__.primary_key:
            pk[key.name] = data[key.name]

        return self.session.get(model, pk)

    def _load(self, transactions):
        """Performs the operations of a group transaction."""
        logger = Logger.get_logger()

        for action in transactions:
            operations = action.prepare()
            with self.session.begin():
                for op in operations:
                    try:
                        if op.type == OperationType.INSERT:
                            obj = op.model(**op.data)
                            self.session.add(obj)
                        elif op.type == OperationType.DELETE:
                            obj = self._get_obj_by_pk(op.model, op.data)
                            self.session.delete(obj)
                        elif op.type == OperationType.UPDATE:
                            obj = self._get_obj_by_pk(op.model, op.data)
                            # due to dictdiff/partial updates we only update those
                            # keys that changed the value. there is no instantiation
                            # of the data class with the op.data so they dont need
                            # values to be defined as optional
                            for key, value in op.data.items():
                                setattr(obj, key, value)

                        self.session.flush()
                    except Exception:
                        logger.exception(
                            f"Could not load {action.data.tx_id} ({action.name})",
                            exc_info=1,
                        )
                        self.session.rollback()
                        raise
                # commit all transaction group or none
                self.session.commit()

    def run(self, entries, cleanup=False):
        """Load entries."""
        self._load(entries)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration extract module."""

from .base import Extract
from .jsonlines import JSONLExtract
from .null import NullExtract
from .transactions import Tx

__all__ = (
    "Extract",
    "JSONLExtract",
    "NullExtract",
    "Tx",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transform interfaces."""

from functools import partial

from .base import Transform


class IdentityTransform(Transform):
    """Transform class to yield the received item without change."""

    def _transform(self, entry):
        """Transform entry."""
        return entry


class IdentityDictKeyMixin:
    """Translates a function name to a dictionary key.

    Prevents unnecessary implementations where the return value of a function
    is merely the dictionary value of key with the same name as the function.
    Or as in this case, prefixed with underscore.

    .. highlight:: python
    .. code-block:: python

        def _test(data):
            return data["test"]
    """

    @staticmethod
    def unprefixed_dict_access(key, data):
        """Removes the prefix from the key and returns the value of the dict."""
        if key.startswith("_"):
            key = key[1:]

        return data[key]

    def __getattr__(self, item):
        """Fallback to the dict access when the attr/func was not found in the class."""
        # note that python will give priority to the implemented functions before calling
        # this method, the one called always is __getattribute__
        # see test_identity_dict_key_mixin for more details
        return partial(self.unprefixed_dict_access, item)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration GitHub table load module."""

from ...load.postgresql.bulk.generators import TableGenerator
from ...state import STATE
from ..models.github import Release


class ReleaseTableGenerator(TableGenerator):
    """GitHub release table load."""

    def __init__(self):
        """Constructor."""
        super().__init__(tables=[Release], pks=[])

    def _generate_rows(self, data, **kwargs):
        recid = data.pop("recid", None)
        release = dict(**data)
        if recid:
            record = STATE.RECORDS.get(str(recid))
            release["record_id"] = record["id"]
        yield Release(**release)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record transform interfaces."""

from abc import abstractmethod

from ...transform import Entry, Transform


class RDMRecordTransform(Transform):
    """Base class for record data transformation."""

    @abstractmethod
    def _parent(self, entry):
        """Transform the parent record."""
        pass

    @abstractmethod
    def _record(self, entry):
        """Transform the record."""
        pass

    @abstractmethod
    def _draft(self, entry):
        """Transform the draft."""
        pass

    @abstractmethod
    def _record_files(self, entry):
        """Transform the record files."""
        pass

    @abstractmethod
    def _draft_files(self, entry):
        """Transform the draft files."""
        pass

    def _transform(self, entry):
        """Transform a single entry."""
        # the functions receive the full record/data entry
        # while in most cases the full view is not needed
        # since this is a low level tool used only by users
        # with deep system knowledge providing the flexibility
        # is future proofing and simplifying the interface
        return {
            "record": self._record(entry),
            "draft": self._draft(entry),
            "parent": self._parent(entry),
            "record_files": self._record_files(entry),
            "draft_files": self._draft_files(entry),
        }


class RDMRecordEntry(Entry):
    """Transform a single record entry."""

    @abstractmethod
    def _id(self, entry):
        """Returns the rdm record uuid."""
        pass

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date of the record."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date of the record."""
        pass

    @abstractmethod
    def _version_id(self, entry):
        """Returns the version id of the record.

        {Elastic,Open}search will fail with values lower than 1.
        """
        pass

    @abstractmethod
    def _index(self, entry):
        """Returns the index of the record."""
        pass

    @abstractmethod
    def _recid(self, entry):
        """Returns the recid of the record."""
        pass

    @abstractmethod
    def _pids(self, entry):
        """Returns the pids of the record."""
        pass

    @abstractmethod
    def _bucket_id(self, entry):
        """Returns the bucket id of the record."""
        pass

    @abstractmethod
    def _media_bucket_id(self, entry):
        """Returns the bucket id of the record."""
        pass

    @abstractmethod
    def _files(self, entry):
        """Transform the files of a record."""
        pass

    @abstractmethod
    def _media_files(self, entry):
        """Transform the media files of a record."""
        pass

    @abstractmethod
    def _access(self, entry):
        """Transform the access of a record."""
        pass

    @abstractmethod
    def _metadata(self, entry):
        """Transform the metadata of a record.

        No schema/dictionary is enforced as choice balanced between
        duplicating the schema with abstract methods and allowing for flexibility
        i.e. different metadata schema.
        """
        pass

    @abstractmethod
    def _custom_fields(self, entry):
        """Transform the custom fields of a record."""
        pass

    def transform(self, entry):
        """Transform a record single entry."""
        return {
            "id": self._id(entry),
            "created": self._created(entry),
            "updated": self._updated(entry),
            "version_id": self._version_id(entry),
            "index": self._index(entry),
            "bucket_id": self._bucket_id(entry),
            "media_bucket_id": self._media_bucket_id(entry),
            "json": {
                "id": self._recid(entry),
                "pids": self._pids(entry),
                "files": self._files(entry),
                "media_files": self._media_files(entry),
                "metadata": self._metadata(entry),
                "access": self._access(entry),
                "custom_fields": self._custom_fields(entry),
            },
        }


class RDMRecordFileEntry(Entry):
    """Transform a single record file entry.

    Connects records with files.
    """

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _json(self, entry):
        """Returns the rdm record file metadata."""
        pass

    @abstractmethod
    def _version_id(self, entry):
        """Returns the rdm record file version id."""
        pass

    @abstractmethod
    def _key(self, entry):
        """Returns the rdm record file key name."""
        pass

    @abstractmethod
    def _object_version_id(self, entry):
        """Returns the associated file object version ID."""
        pass

    def transform(self, entry):
        """Transform a single record files entry."""
        return {
            "created": self._created(entry),
            "updated": self._updated(entry),
            "json": self._json(entry),
            "version_id": self._version_id(entry),
            "key": self._key(entry),
            "object_version_id": self._object_version_id(entry),
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Requests table generator tests."""

from invenio_rdm_migrator.streams.files.load import FilesTableGenerator
from invenio_rdm_migrator.streams.models.files import (
    FilesBucket,
    FilesInstance,
    FilesObjectVersion,
)


def test_generate_rows(transformed_files_entry):
    """Test the row generation of the request table generator."""
    tg = FilesTableGenerator()
    rows = list(tg._generate_rows(transformed_files_entry))
    expected_rows = [
        FilesInstance(
            id="3",
            created="2023-04-19",
            updated="2023-04-19",
            uri="path/to/file",
            storage_class="L",
            size=1234,
            checksum="md5:abcd1234",
            readable=True,
            writable=True,
            last_check_at=None,
            last_check=False,
        ),
        FilesBucket(
            id="1",
            created="2023-04-19",
            updated="2023-04-19",
            default_location="1",
            default_storage_class="L",
            size=1234,
            quota_size=None,
            max_file_size=None,
            locked=False,
            deleted=False,
        ),
        FilesObjectVersion(
            version_id="2",
            created="2023-04-19",
            updated="2023-04-19",
            key="file.txt",
            bucket_id="1",
            file_id="3",
            _mimetype=None,
            is_head=True,
        ),
    ]

    assert rows == expected_rows



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Logging module."""

import logging
import sys

from .utils import ts


class Logger:
    """Migrator logger."""

    @classmethod
    def initialize(cls, log_dir):
        """Constructor."""
        formatter = logging.Formatter(
            fmt="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        logger = logging.getLogger("migrator")
        logger.setLevel(logging.INFO)
        # errors to file
        fh = logging.FileHandler(log_dir / "error.log")
        fh.setLevel(logging.ERROR)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        # info to stream/stdout
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        sh.setLevel(logging.INFO)
        logger.addHandler(sh)

    @classmethod
    def get_logger(cls):
        """Get migration logger."""
        return logging.getLogger("migrator")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""User actions module."""

from dataclasses import dataclass
from typing import Optional

from ....actions import LoadAction, LoadData
from ....load.postgresql.transactions.operations import Operation, OperationType
from ....transform import EncryptMixin
from ...models.users import LoginInformation, SessionActivity, User

# GLOBAL FIXME: partial updates would allow to remove the encrypt mixin from many of the
# actions, since only user registration and edit (with password change should) do the
# re-encryption. Would save time too.


@dataclass
class UserData(LoadData):
    """User action data."""

    user: dict
    login_information: Optional[dict] = None
    sessions: Optional[list] = None


class UserRegistrationAction(LoadAction, EncryptMixin):
    """Registers a user."""

    name = "register-user"
    data_cls = UserData

    def __init__(self, data, **kwargs):
        """Constructor."""
        # Explicit calls, otherwise MRO only initializes LoadAction
        LoadAction.__init__(self, data, **kwargs)
        EncryptMixin.__init__(self)

    def _generate_rows(self, **kwargs):
        """Generates rows for a new user."""
        self.data.user["password"] = self.re_encrypt(self.data.user["password"])

        yield Operation(OperationType.INSERT, User, self.data.user)

        if self.data.login_information:
            yield Operation(
                OperationType.INSERT,
                LoginInformation,
                dict(user_id=self.data.user["id"], **self.data.login_information),
            )


# Handles many types of changes:
# - Profile change (email, username, full name)
# - Password change (including re-encryption)
# - User confirmation
# - User deactivation
class UserEditAction(LoadAction, EncryptMixin):
    """Registers a user."""

    name = "edit-user"
    data_cls = UserData

    def __init__(self, data, **kwargs):
        """Constructor."""
        # Explicit calls, otherwise MRO only initializes LoadAction
        LoadAction.__init__(self, data, **kwargs)
        EncryptMixin.__init__(self)

    def _generate_rows(self, **kwargs):
        """Generates rows for a user edit."""
        self.data.user["password"] = self.re_encrypt(self.data.user["password"])

        yield Operation(OperationType.UPDATE, User, self.data.user)

        if self.data.login_information:
            yield Operation(
                OperationType.UPDATE,
                LoginInformation,
                dict(user_id=self.data.user["id"], **self.data.login_information),
            )


class UserProfileEditAction(LoadAction):
    """Registers a user."""

    name = "edit-user-profile"

    def __init__(
        self,
        tx_id,
        profile,
    ):
        """Constructor."""
        super().__init__(tx_id)
        assert profile  # i.e. not None as parameter
        self.profile = profile

    def _generate_rows(self, **kwargs):
        """Generates rows for a new draft."""
        pass


class UserDeactivationAction(LoadAction, EncryptMixin):
    """Deactivate a user.

    For example, flag it as spam.
    """

    name = "deactivate-user"
    data_cls = UserData

    def __init__(self, data, **kwargs):
        """Constructor."""
        # Explicit calls, otherwise MRO only initializes LoadAction
        LoadAction.__init__(self, data, **kwargs)
        EncryptMixin.__init__(self)

    def _generate_rows(self, **kwargs):
        """Generates rows for a new draft."""
        assert not self.data.user["active"]

        self.data.user["password"] = self.re_encrypt(self.data.user["password"])

        yield Operation(OperationType.UPDATE, User, self.data.user)

        if self.data.login_information:
            yield Operation(
                OperationType.UPDATE,
                LoginInformation,
                dict(user_id=self.data.user["id"], **self.data.login_information),
            )

        for session in self.data.sessions:
            yield Operation(OperationType.DELETE, SessionActivity, session)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Files actions module."""

from copy import deepcopy
from dataclasses import dataclass

from ....actions import LoadAction, LoadData
from ....load.ids import generate_uuid
from ....load.postgresql.transactions.operations import Operation, OperationType
from ....state import STATE
from ...models.files import FilesBucket, FilesInstance, FilesObjectVersion
from ...models.records import RDMDraftFile


@dataclass
class FileUploadData(LoadData):
    """File upload action data."""

    bucket: dict
    object_version: dict
    file_instance: dict
    file_record: dict


class DraftFileUploadAction(LoadAction):
    """RDM draft creation."""

    name = "draft-upload-file"
    data_cls = FileUploadData
    pks = [("file_record", "id", generate_uuid)]

    def _generate_rows(self, **kwargs):
        """Generates rows for a new draft."""
        # if we were to use the state for consistency checks
        # the bucket should already exist
        yield Operation(OperationType.UPDATE, FilesBucket, self.data.bucket)
        yield Operation(
            OperationType.INSERT, FilesObjectVersion, self.data.object_version
        )
        yield Operation(OperationType.INSERT, FilesInstance, self.data.file_instance)

        cached_bucket = STATE.BUCKETS.get(self.data.bucket["id"])
        self.data.file_record["record_id"] = cached_bucket["draft_id"]

        _cache_fr_data = deepcopy(self.data.file_record)
        STATE.FILE_RECORDS.add(_cache_fr_data.pop("id"), _cache_fr_data)

        yield Operation(OperationType.INSERT, RDMDraftFile, self.data.file_record)
